# -*- coding: utf-8 -*-
#
# This file inherits code from original editor.py from NINJA-IDE (http://ninja-ide.org).
# and https://nachtimwald.com/2009/08/15/qtextedit-with-line-numbers/

from __future__ import absolute_import
from __future__ import unicode_literals

import re
import sys
import os
import traceback
from PyQt4.QtGui import QPlainTextEdit,QFrame,QPainter
from PyQt4.QtGui import QTextCursor,QWidget,QHBoxLayout
from PyQt4.QtCore import Qt,QRect
from PyQt4.QtCore import SIGNAL
from PyQt4 import QtGui,QtCore

endCharsForIndent = [':', '{', '(', '[']

patIndent = re.compile('^\s+')

BRACE_DICT = {')': '(', ']': '[', '}': '{', '(': ')', '[': ']', '{': '}'}

BRACES = {'{': '}',
          '[': ']',
          '(': ')'}
QUOTES = {'"': '"',
          "'": "'"}

USE_TABS = False
ALLOW_WORD_WRAP = False
INDENT = 4
# by default Unix (\n) is used
USE_PLATFORM_END_OF_LINE = False
MARGIN_LINE = 80
SHOW_MARGIN_LINE = True
REMOVE_TRAILING_SPACES = True
SHOW_TABS_AND_SPACES = True

def get_indentation(line, indent=INDENT, useTabs=USE_TABS):
    global patIndent
    global endCharsForIndent
    indentation = ''
    if len(line) > 0 and line[-1] in endCharsForIndent:
        if useTabs:
            indentation = '\t'
        else:
            indentation = ' ' * indent
    elif len(line) > 0 and line[-1] == ',':
        count = [x for x in endCharsForIndent[1:]
            if (line.count(x) - line.count(closeBraces[x])) % 2 != 0]
        if count:
            if useTabs:
                indentation = '\t'
            else:
                indentation = ' ' * indent
    space = patIndent.match(line)
    if space is not None:
        return space.group() + indentation
    return indentation


class PyEditor(QPlainTextEdit):

    def __init__(self):
        QPlainTextEdit.__init__(self)
        self.indent = INDENT
        self.useTabs = USE_TABS
        # GOTO LINE shortcut
        QtGui.QShortcut(QtGui.QKeySequence("Ctrl+L"), self, self._scroll_To_Line, context=Qt.WidgetShortcut)
        #Dict functions for KeyPress
        self.preKeyPress = {
            Qt.Key_Tab: self.__insert_indentation,
            Qt.Key_Backspace: self.__backspace,
            Qt.Key_Home: self.__home_pressed,
            Qt.Key_Enter: self.__ignore_extended_line,
            Qt.Key_Return: self.__ignore_extended_line,
            Qt.Key_BracketRight: self.__brace_completion,
            Qt.Key_BraceRight: self.__brace_completion,
            Qt.Key_ParenRight: self.__brace_completion,
            Qt.Key_Apostrophe: self.__quot_completion,
            Qt.Key_QuoteDbl: self.__quot_completion}

        self.postKeyPress = {
            Qt.Key_Enter: self.__auto_indent,
            Qt.Key_Return: self.__auto_indent,
            Qt.Key_BracketLeft: self.__complete_braces,
            Qt.Key_BraceLeft: self.__complete_braces,
            Qt.Key_ParenLeft: self.__complete_braces,
            Qt.Key_Apostrophe: self.__complete_quotes,
            Qt.Key_QuoteDbl: self.__complete_quotes}


    def _scroll_To_Line(self):
        line,ok = QtGui.QInputDialog.getInt(self,"Go to line","Enter line number")
        if not ok or (line <= 0):
            return
        self.moveCursor(QtGui.QTextCursor.Start)
        cursor = QtGui.QTextCursor(self.document().findBlockByLineNumber(line-1))
        cursor.select(QtGui.QTextCursor.LineUnderCursor)
        self.setTextCursor(cursor)


    def _add_line_increment(self, lines, blockModified, diference):
        def _inner_increment(line):
            if line < blockModified:
                return line
            return line + diference
        return list(map(_inner_increment, lines))

    def _add_line_increment_for_dict(self, data, blockModified, diference):
        def _inner_increment(line):
            if line < blockModified:
                return line
            newLine = line + diference
            summary = data.pop(line)
            data[newLine] = summary
            return newLine
        list(map(_inner_increment, list(data.keys())))
        return data


    def __insert_indentation(self, event):
        if self.textCursor().hasSelection():
            self.indent_more()
        elif self.useTabs:
            return False
        else:
            self.textCursor().insertText(' ' * self.indent)
        return True

    def __backspace(self, event):
        if self.textCursor().hasSelection() or self.useTabs:
            return False
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.StartOfLine, QTextCursor.KeepAnchor)
        text = str(cursor.selection().toPlainText())
        if (len(text) % self.indent == 0) and text.isspace():
            cursor.movePosition(QTextCursor.StartOfLine)
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor,
                self.indent)
            cursor.removeSelectedText()
            return True

    def __home_pressed(self, event):
        if event.modifiers() == Qt.ControlModifier:
            return False
        elif event.modifiers() == Qt.ShiftModifier:
            move = QTextCursor.KeepAnchor
        else:
            move = QTextCursor.MoveAnchor
        if self.textCursor().atBlockStart():
            self.moveCursor(QTextCursor.WordRight, move)
            return True

        cursor = self.textCursor()
        position = cursor.position()
        self.moveCursor(QTextCursor.StartOfLine, move)
        self.moveCursor(QTextCursor.WordRight, move)
        if position != self.textCursor().position() and \
           str(cursor.block().text()).startswith(' '):
            return True

    def __ignore_extended_line(self, event):
        if event.modifiers() == Qt.ShiftModifier:
            return True

    def __set_selection_from_pair(self, begin, end):
        """Set the current editor cursor with a selection from a given pair of
        positions"""
        cursor = self.textCursor()
        cursor.setPosition(begin)
        cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def __reverse_select_text_portion_from_offset(self, begin, end):
        """Backwards select text, go from current+begin to current - end
        possition, returns text"""
        cursor = self.textCursor()
        cursor_position = cursor.position()
        cursor.setPosition(cursor_position + begin)
        #QT silently fails on invalid position, ergo breaks when EOF < begin
        while (cursor.position() == cursor_position) and begin > 0:
            begin -= 1
            cursor.setPosition(cursor_position + begin)
        cursor.setPosition(cursor_position - end, QTextCursor.KeepAnchor)
        selected_text = cursor.selectedText()
        return selected_text

    def __quot_completion(self, event):
        """Indicate if this is some sort of quote that needs to be completed
        This is a very simple boolean table, given that quotes are a
        simmetrical symbol, is a little more cumbersome guessing the completion
        table.
        """
        text = event.text()
        pos = self.textCursor().position()
        next_char = str(self.get_selection(pos, pos + 1)).strip()
        if self.cursor_inside_string() and text == next_char:
            self.moveCursor(QTextCursor.Right)
            return True
        PENTA_Q = 5 * text
        TETRA_Q = 4 * text
        TRIPLE_Q = 3 * text
        DOUBLE_Q = 2 * text
        supress_echo = False
        pre_context = self.__reverse_select_text_portion_from_offset(0, 3)
        pos_context = self.__reverse_select_text_portion_from_offset(3, 0)
        if pre_context == pos_context == TRIPLE_Q:
            supress_echo = True
        elif pos_context[:2] == DOUBLE_Q:
            pre_context = self.__reverse_select_text_portion_from_offset(0, 4)
            if pre_context == TETRA_Q:
                supress_echo = True
        elif pos_context[:1] == text:
            pre_context = self.__reverse_select_text_portion_from_offset(0, 5)
            if pre_context == PENTA_Q:
                supress_echo = True
            elif pre_context[-1] == text:
                supress_echo = True
        if supress_echo:
            self.moveCursor(QTextCursor.Right)
        return supress_echo

    def __brace_completion(self, event):
        """Indicate if this symbol is part of a given pair and needs to be
        completed.
        """
        text = event.text()
        if text in list(BRACES.values()):
            portion = self.__reverse_select_text_portion_from_offset(1, 1)
            brace_open = portion[0]
            brace_close = (len(portion) > 1) and portion[1] or None
            balance = BRACE_DICT.get(brace_open, None) == text == brace_close
            if balance:
                self.moveCursor(QTextCursor.Right)
                return True

    def __auto_indent(self, event):
        text = self.textCursor().block().previous().text()
        spaces = get_indentation(text, self.indent, self.useTabs)
        self.textCursor().insertText(spaces)
        if text != '' and text == ' ' * len(text):
            self.moveCursor(QTextCursor.Up)
            self.moveCursor(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
            self.textCursor().removeSelectedText()
            self.moveCursor(QTextCursor.Down)
        cursor = self.textCursor()
        cursor.setPosition(cursor.position())
        self.setTextCursor(cursor)

    def insert_new_line(self):
        cursor = self.textCursor()
        at_block_end = cursor.atBlockEnd()
        cursor.movePosition(QTextCursor.EndOfLine)
        cursor.insertBlock()
        if not at_block_end:
            self.moveCursor(QTextCursor.Down)
        self.__auto_indent(None)

    def __complete_braces(self, event):
        """Complete () [] and {} using a mild inteligence to see if corresponds
        and also do some more magic such as complete in classes and functions.
        """
        brace = event.text()
        if brace not in BRACES:
            # Thou shalt not waste cpu cycles if this brace compleion dissabled
            return
        text = self.textCursor().block().text()
        complementary_brace = BRACE_DICT.get(brace)
        token_buffer = []
        _, tokens = self.__tokenize_text(text)
        is_unbalance = 0
        for tkn_type, tkn_rep, tkn_begin, tkn_end in tokens:
            if tkn_rep == brace:
                is_unbalance += 1
            elif tkn_rep == complementary_brace:
                is_unbalance -= 1
            if tkn_rep.strip() != "":
                token_buffer.append((tkn_rep, tkn_end[1]))
            is_unbalance = (is_unbalance >= 0) and is_unbalance or 0

        if (self.lang == "python") and (len(token_buffer) == 3) and \
                (token_buffer[2][0] == brace) and (token_buffer[0][0] in
                                                        ("def", "class")):
            #are we in presence of a function?
            self.textCursor().insertText("):")
            self.__fancyMoveCursor(QTextCursor.Left, 2)
            self.textCursor().insertText(self.selected_text)
        elif token_buffer and (not is_unbalance) and \
           self.selected_text:
            self.textCursor().insertText(self.selected_text)
        elif is_unbalance:
            pos = self.textCursor().position()
            next_char = self.get_selection(pos, pos + 1).strip()
            if self.selected_text or next_char == "":
                self.textCursor().insertText(complementary_brace)
                self.moveCursor(QTextCursor.Left)
                self.textCursor().insertText(self.selected_text)

    def __complete_quotes(self, event):
        """
        Completion for single and double quotes, which since are simmetrical
        symbols used for different things can not be balanced as easily as
        braces or equivalent.
        """
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.StartOfLine,
            QTextCursor.KeepAnchor)
        symbol = event.text()
        if symbol in QUOTES:
            pre_context = self.__reverse_select_text_portion_from_offset(0, 3)
            if pre_context == 3 * symbol:
                self.textCursor().insertText(3 * symbol)
                self.__fancyMoveCursor(QTextCursor.Left, 3)
            else:
                self.textCursor().insertText(symbol)
                self.moveCursor(QTextCursor.Left)
            self.textCursor().insertText(self.selected_text)

    def keyPressEvent(self, event):
        #On Return == True stop the execution of this method
        if self.preKeyPress.get(event.key(), lambda x: False)(event):
            #emit a signal then plugings can do something
            self.emit(SIGNAL("keyPressEvent(QEvent)"), event)
            return
        self.selected_text = self.textCursor().selectedText()

        QPlainTextEdit.keyPressEvent(self, event)

        self.postKeyPress.get(event.key(), lambda x: False)(event)

        #emit a signal then plugings can do something
        self.emit(SIGNAL("keyPressEvent(QEvent)"), event)


class PyEditWidget(QFrame):

    class NumberBar(QWidget):

        def __init__(self, *args):
            QWidget.__init__(self, *args)
            self.edit = None
            # This is used to update the width of the control.
            # It is the highest line that is currently visibile.
            self.highest_line = 0
            self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
            

        def setTextEdit(self, edit):
            self.edit = edit

        def update(self, *args):
            '''
            Updates the number bar to display the current set of numbers.
            Also, adjusts the width of the number bar if necessary.
            '''
            # The + 4 is used to compensate for the current line being bold.
            width = self.fontMetrics().width(str(self.highest_line)) + 10
            if self.width() != width:
                self.setFixedWidth(width)
            QWidget.update(self, *args)

        def paintEvent(self, event):
            font_metrics = self.fontMetrics()
            current_line = self.edit.document().findBlock(self.edit.textCursor().position()).blockNumber() + 1

            painter = QPainter(self)

            line_count = 0
            # Iterate over all text blocks in the document.
            block = self.edit.document().begin()

            while block.isValid():
                line_count += 1
               
                block_top = self.edit.blockBoundingGeometry(block).translated(self.edit.contentOffset()).top()

                # We want the line number for the selected line to be bold.
                bold = False
                if line_count == current_line:
                    bold = True
                    font = painter.font()
                    font.setBold(True)
                    painter.setFont(font)

                # Draw the line number right justified at the y position of the
                # line. 3 is a magic padding number. drawText(x, y, text).
                paint_rect = QRect(0, block_top, self.width(), font_metrics.height())
                painter.drawText(paint_rect, Qt.AlignLeft, unicode(line_count))

                # Remove the bold style if it was set previously.
                if bold:
                    font = painter.font()
                    font.setBold(False)
                    painter.setFont(font)

                block = block.next()

            self.highest_line = line_count
            painter.end()

            QWidget.paintEvent(self, event)


    def _loadScript(self):
        try:
            fileName = QtGui.QFileDialog.getOpenFileName(None,'Load script file', self.script_path, filter='*.py')
            if fileName:
                f = open(fileName)
                self.edit.setPlainText(f.read())
                f.close()
                self.fileName = fileName
                self.saveButton.setText("Save")
                if self.main_gui is not None:
                    self.main_gui.setWindowTitle("Donki Scripting - %s" % self.fileName)
                self.last_saved_text = self.edit.toPlainText()
                self.text_changed = False
        except:
            self._errorMessage(traceback.format_exc())

    def _saveScript(self, forced = False):
        try:
            if self.edit.toPlainText() == '':
                return
            if self.fileName is not None:
                suggested = self.fileName
            else:
                suggested = self.script_path
            if not forced:
                opts = QtGui.QFileDialog.DontConfirmOverwrite
                fileName = QtGui.QFileDialog.getSaveFileName(self,'Save script', suggested, filter='*.py', options=opts)
                if fileName:
                    if os.path.exists(fileName):
                        qm = QtGui.QMessageBox
                        ret = qm.question(self,'File exists', "File exists: do you want to overwrite it?", qm.Yes | qm.No)
                        if ret != qm.Yes:
                            return
            else:
                fileName = self.fileName
            self.fileName = fileName
            if fileName:
                f = open(fileName,"w")
                f.write(self.edit.toPlainText())
                f.close()
            self.saveButton.setText("Save *")
            if self.main_gui is not None:
                self.main_gui.setWindowTitle("Donki Scripting - %s" % self.fileName)
            self.last_saved_text = self.edit.toPlainText()
            self.text_changed = False
        except:
            self._errorMessage(traceback.format_exc())

    def _textChanged(self):
        try:
            if self.last_saved_text != self.edit.toPlainText():
                self.text_changed = True
                self.saveButton.setText("Save *")
                if self.main_gui is not None:
                    self.main_gui.setWindowTitle("Donki Scripting - %s *" % self.fileName)
            else:
                self.text_changed = False
                self.saveButton.setText("Save")
        except:
            traceback.print_exc()
            

    def _errorMessage(self, err_message):
            msg = QtGui.QMessageBox()
            msg.setIcon(QtGui.QMessageBox.Warning)
            msg.setText(err_message)
            msg.setWindowTitle("Error")
            msg.setStandardButtons(QtGui.QMessageBox.Ok)
            retval = msg.exec_()


    def __init__(self, main_gui, default_path = "."):
        QFrame.__init__(self)

        self.fileName = None
        self.text_changed = False
        
        self.main_gui = main_gui
        self.script_path = default_path

        self.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)

        self.edit = PyEditor()
        self.edit.setFrameStyle(QFrame.NoFrame)
        #
        self.number_bar = self.NumberBar()
        self.number_bar.setTextEdit(self.edit)
        #
        self.saveButton = QtGui.QPushButton(self)
        self.saveButton.setAutoDefault(False)
        self.saveButton.setObjectName("saveButton")
        self.saveButton.setText("Save")
        self.saveButton.clicked.connect(self._saveScript)
        #
        self.loadButton = QtGui.QPushButton(self)
        self.loadButton.setAutoDefault(False)
        self.loadButton.setObjectName("loadButton")
        self.loadButton.setText("Load")
        self.loadButton.clicked.connect(self._loadScript)

        #
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.bbox = QtGui.QGridLayout(self)
        self.bbox.setSpacing(0)
        self.bbox.setMargin(0)
        self.bbox.addWidget(self.loadButton,0,0,1,1)
        self.bbox.addWidget(self.saveButton,0,1,1,1)
        self.bbox.addItem(spacerItem,0,2,1,1)
        #
        self.hbox = QHBoxLayout()
        self.hbox.setSpacing(0)
        self.hbox.setMargin(0)
        self.hbox.addWidget(self.number_bar)
        self.hbox.addWidget(self.edit)
        #
        self.bbox.addLayout(self.hbox,1,0,1,3)
        #
        self.edit.installEventFilter(self)
        self.edit.viewport().installEventFilter(self)
        #
        # Add some shortcuts
        QtGui.QShortcut(QtGui.QKeySequence("Ctrl+S"), self.edit, self._saveScript, context=QtCore.Qt.WidgetShortcut)
        self.edit.textChanged.connect(self._textChanged)
        self.last_saved_text = self.edit.toPlainText()


    def eventFilter(self, object, event):
        # Update the line numbers for all events on the text edit and the viewport.
        # This is easier than connecting all necessary singals.
        if object in (self.edit, self.edit.viewport()):
            self.number_bar.update()
            return False
        return QFrame.eventFilter(object, event)

    def getTextEdit(self):
        return self.edit
